from dataclasses import dataclass, field
from typing import Optional
from datetime import datetime
from typing import Type
from copy import deepcopy

from nema.connectivity import ConnectivityManager

from .data_properties import DataProperties


@dataclass
class CacheInfo:
    last_time_updated: datetime = datetime(1970, 1, 1)
    number_of_times_hit: int = 0
    in_sync_cloud: bool = False


@dataclass
class Data:
    _global_id: str
    _data: DataProperties
    _cache_info: CacheInfo = field(default_factory=CacheInfo)
    _userdefined_id: Optional[str] = None

    @classmethod
    def init_from_cloud(
        cls,
        global_id: str,
        data_class: Type[DataProperties],
        userdefined_id: Optional[str] = None,
    ):
        data = data_class()

        return cls(
            _global_id=global_id,
            _userdefined_id=userdefined_id,
            _data=data,
        )

    @property
    def global_id(self):
        return self._global_id

    @property
    def data(self):
        # update cache info
        self._cache_info.number_of_times_hit += 1

        return self._data

    @property
    def value(self):
        return self.data.get_value()

    def update_data(self, new_data: DataProperties):
        if self._cache_info.number_of_times_hit > 0:
            raise ValueError("Data can only be updated if it has not been accessed")

        self._data = new_data
        self._cache_info.last_time_updated = datetime.now()
        self._cache_info.number_of_times_hit = 0
        self._cache_info.in_sync_cloud = False

    @property
    def userdefined_id(self):
        return self._userdefined_id

    @userdefined_id.setter
    def userdefined_id(self, value: str):
        self._userdefined_id = value

    @property
    def cache_info(self):
        return deepcopy(self._cache_info)

    def sync_from_API(self):
        # Sync data from the API

        conn = ConnectivityManager()

        raw_data = conn.pull_in_memory_data(self.global_id)

        # we need to use _data here, otherwise we record a cache hit
        self._data = self._data.__nema_unmarshall__(raw_data["data_properties"])

    def marshall_data_properties(self):
        return self._data.__nema_marshall__()

    @property
    def is_updated(self):
        return (
            self._cache_info.in_sync_cloud == False
            and self._cache_info.last_time_updated.year > 1980
        )

    @property
    def data_type(self):
        return self._data.data_type

    def process_output_data(self, destination_folder: str):

        core_dict = {
            "global_id": self.global_id,
            "data_properties": self.marshall_data_properties(),
            "data_type": self.data_type,
        }

        if self._data.is_blob_data:
            core_dict["file_name"] = self._data.write_data_to_file_and_return_file_name(
                destination_folder=destination_folder
            )

        return core_dict



from dataclasses import dataclass
import os
import json
import requests
import mimetypes


from nema.utils.global_config import GLOBAL_CONFIG


def get_mime_type(file_path):
    mime_type, _ = mimetypes.guess_type(file_path)
    return mime_type or "application/octet-stream"


NEMA_PATH = os.path.expanduser("~/.nema")
AUTH_PATH = os.path.join(NEMA_PATH, "auth")


def save_auth_data(refresh_token: str, access_token: str):
    os.makedirs(os.path.dirname(AUTH_PATH), exist_ok=True)

    with open(AUTH_PATH, "w") as f:
        f.write(
            json.dumps({"refresh_token": refresh_token, "access_token": access_token})
        )


@dataclass
class ConnectivityManager:
    _auth_token: str = ""
    _refresh_token: str = ""

    def __post_init__(self):

        if os.path.exists(AUTH_PATH):
            with open(AUTH_PATH, "r") as f:
                data = json.load(f)
                self._auth_token = data["access_token"]
                self._refresh_token = data["refresh_token"]

    @property
    def auth_token(self):
        return self._auth_token

    def get_headers(self):
        return {"Authorization": f"Bearer {self.auth_token}"}

    def refresh_auth_token(self):
        """Refresh the authentication token"""

        tenant_url = GLOBAL_CONFIG.tenant_api_url
        login_url = f"{tenant_url}/authentication/refresh-token"

        response = requests.post(login_url, json={"token": self._refresh_token})

        if not response.ok:
            if response.status_code == 401:
                raise ConnectionError("Token refresh failed, please login again.")
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        response_data = response.json()
        self._refresh_token = response_data["refresh_token"]
        self._auth_token = response_data["access_token"]

        save_auth_data(self._refresh_token, self._auth_token)

    def pull_in_memory_data(self, global_id: int):

        data_api_url = f"{GLOBAL_CONFIG.project_api_url}/artifacts/data/{global_id}"

        # Make a request to the login URL
        response = requests.get(
            data_api_url,
            headers=self.get_headers(),
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.pull_in_memory_data(global_id)
            else:
                raise ConnectionError(
                    "Connection error: {response.status_code} -- {response.text}"
                )

        return response.json()

    def push_workflow_update(
        self, global_id: int, workflow_raw_data: dict, files: list[str] = []
    ):

        url = f"{GLOBAL_CONFIG.project_api_url}/artifacts/workflows/{global_id}/upload-external-results"

        response = requests.post(
            url,
            headers={**self.get_headers()},
            data=workflow_raw_data,
            files=[
                ("file", (fp.split("/")[-1], open(fp, "rb"), get_mime_type(fp)))
                for fp in files
            ],
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.push_workflow_update(global_id, workflow_raw_data)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        return response.json()

    def create_workflow(self, workflow_data: dict):
        """Create a new workflow in the nema platform"""

        url = f"{GLOBAL_CONFIG.project_api_url}/artifacts/workflows"

        response = requests.post(
            url,
            headers={**self.get_headers(), "Content-Type": "application/json"},
            json=workflow_data,
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.create_workflow(workflow_data)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        return_data = response.json()
        return return_data["created_global_id"]



from dataclasses import dataclass, field
import shutil
import tempfile
import os

from nema.utils.file_name import generate_random_file_name


@dataclass
class DataProperties:

    @property
    def is_blob_data(self):
        return False

    @property
    def data_type(self):
        raise NotImplementedError("This method should be implemented by the subclass")

    def get_value(self):
        raise NotImplementedError("This method should be implemented by the subclass")

    def __nema_marshall__(self):
        raise NotImplementedError("This method should be implemented by the subclass")

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        raise NotImplementedError("This method should be implemented by the subclass")


@dataclass
class SingleValue(DataProperties):

    def __nema_marshall__(self):
        return {"value": self.value}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(value=data["value"])

    def get_value(self):
        return self.value


@dataclass
class BooleanValue(SingleValue):
    value: bool = False

    @property
    def data_type(self):
        return "BOOL"


@dataclass
class StringValue(SingleValue):
    value: str = ""

    @property
    def data_type(self):
        return "STRING"


@dataclass
class IntegerValue(SingleValue):
    value: int = 0

    @property
    def data_type(self):
        return "INT"


@dataclass
class FloatValue(SingleValue):
    value: float = 0.0

    @property
    def data_type(self):
        return "FLOAT"


@dataclass
class BlobDataProperties(DataProperties):

    @property
    def is_blob_data(self):
        return True


@dataclass
class FileDataProperties(BlobDataProperties):
    _file_name: str = ""
    _temp_file: tempfile.NamedTemporaryFile = field(init=False, default=None)
    temp_file_name: str = field(init=False, default=None)

    def __nema_marshall__(self):
        return {"file_name": self._file_name}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(_file_name=data["file_name"])

    def __enter__(self):
        self._temp_file = tempfile.NamedTemporaryFile(mode="w+b", delete=False)
        self.temp_file_name = self._temp_file.name
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._temp_file and not self._temp_file.closed:
            self._temp_file.flush()
            self._temp_file.close()

    @property
    def file_extension(self):
        return self._file_name.split(".")[-1]

    def write_data_to_file_and_return_file_name(self, destination_folder: str):
        # move to destination folder
        output_file_name = generate_random_file_name(self.file_extension)
        destination_file_path = os.path.join(destination_folder, output_file_name)
        shutil.move(self.temp_file_name, destination_file_path)

        return output_file_name
